module pfnsLosAlamosModel

! COMMENTS:
!
! Originally coded within RIPL-3 project by
! P. Talou, Group T-2, Los Alamos National Laboratory
! Mail Stop B243, email address talou@lanl.gov
!
! Modifications by:
! R. Capote, IAEA Nuclear Data Section
! email address R.CapoteNoy@iaea.org; rcapotenoy@yahoo.com
!
!---------------------------------------------------------------
!
! RELEASE(S):
!! Date         Release #          Author(s)           Comments
! ----         ---------          ---------           --------
! 27/10/2006      1.0             P. Talou         Original version
! 20/10/2009      1.1             R.Capote          RIPL-3 version
!---------------------------------------------------------------
!
! AUTHOR(S) INFORMATIONS:
!
!-- Tested under Mac OS X 10.5 (Leopard)
!-- with Intel Fortran 95 compiler IFORT (options -O0 and -O3)
!-- with GNU Fortran 90 compiler gfortran (options -O0 and -O3)
!
!-- Tested in PC (c) under Windows XP
!-- with Microsoft Fortran Power Station 
!---------------------------------------------------------------------------
!--
!-- This code computes characteristics of the prompt fission neutrons 
!-- emitted from a given excited compound nucleus, and for a given most
!-- probable fragmentation (light,heavy), within the Madland-Nix (or Los 
!-- Alamos) model, as described in [MN82]:
!--
!-- D.G.Madland and J.R.Nix, "New Calculation of Prompt Fission Neutron
!-- Spectra and Average Prompt Neutron Multiplicities", Nucl. Sci. Eng. 81,
!-- 213-271 (1982).
!--
!---------------------------------------------------------------------------

!-- module ripl2interface used to retrieve nuclear masses from RIPL2 library
use ripl2Masses

  implicit none

  !---------------------------- << INTERFACE >> -------------------------------

  private

  public :: pfnsCalc   !-- [subroutine]
  public :: pfnsType   !-- [derived type]
  public :: pfns       !-- instance of pfnsType

  !---------------------------------------------------------------------------

  !-- set up accuracy of computation with 'rk' [Real Kind]
  integer, parameter :: rk = selected_real_kind (15,307)
  integer, parameter :: r4 = selected_real_kind (6,37)

  !-- PATH to data directory
  character(len=*), parameter :: dataDir = "data/"
  
  !-- main input file
  character(len=*), parameter :: inputFile = "input.pfns"

  !-- main output file
  character(len=*), parameter :: outputFile = "output.pfns"
  
  !-- inverse compound formation cross section file
  character(len=*), parameter :: xsfile = trim(dataDir)//"xs.pfns"

  !---------------------------------------------------------------------------

  !------------------------- << PARAMETERS >> --------------------------------
  
  !-- number of components in multiple-chance fission (default: 5)
  integer, parameter :: Nfc = 5

  !-- number of points used in quadrature integrations
  integer, parameter :: NLeg = 32 !-- number of Gauss-Legendre points
  integer, parameter :: NLag = 32 !-- number of Gauss-Laguerre points

  !------------------------ << DERIVED TYPES >> ------------------------------

  !---------------------------------------------------------------------------
  !-- PUBLIC Type to encapsulate all results from the module
  !---------------------------------------------------------------------------
  !-- numEout                   number of points on outgoing energy grid
  !-- numChanceFission          number of multiple-chance fission
  !-- Eout(:)                   outgoing energy grid (MeV)
  !-- labTotalSpectrum(:)       lab. total spectrum (1/MeV)
  !-- cmTotalSpectrum(:)        cm total spectrum (1/MeV)
  !-- labChanceSpectrum(:,:)    lab spectrum for each chance fission component
  !-- cmChanceSpectrum(:,:)     cm spectrum for each chance fission component
  !-- aveEcm                    average outgoing energy in cm frame (MeV)
  !-- aveElab                   average outgoing energy in lab frame (MeV)
  !-- aveChanceEcm(:)           ave. cm energy for each chance fission (MeV)
  !-- aveChanceElab(:)          ave. lab energy for each chance fission (MeV)
  !-- totalNubar                total neutron multiplicity
  !-- chanceNubar(:)            multiplicity for each chance-fission component
  !---------------------------------------------------------------------------
  type pfnsType
     integer                           :: numEout
     integer                           :: numChanceFission
     real(rk), dimension(:), pointer   :: Eout
     real(rk), dimension(:), pointer   :: labTotalSpectrum
     real(rk), dimension(:), pointer   :: cmTotalSpectrum
     real(rk), dimension(:,:), pointer :: labChanceSpectrum
     real(rk), dimension(:,:), pointer :: cmChanceSpectrum
     real(rk)                          :: aveEcm
     real(rk)                          :: aveElab
     real(rk), dimension(:), pointer   :: aveChanceEcm
     real(rk), dimension(:), pointer   :: aveChanceElab
     real(rk)                          :: totalNubar
     real(rk), dimension(:), pointer   :: chanceNubar
  end type pfnsType

  !---------------------------------------------------------------------------
  !-- Representation of a NUCLEUS
  !---------------------------------------------------------------------------
  type nucleus
     integer  :: Z, A  !-- charge and mass numbers
     integer  :: ZAID  !-- 1000*Z+A
     real(rk) :: alev  !-- level density parameter
     real(rk) :: mass  !-- ground-state mass (MeV)
     real(rk) :: Bn    !-- neutron binding energy (MeV)
     real(rk) :: KE    !-- average kinetic energy per nucleon (MeV/A)
     real(rk) :: Tmax  !-- max. temperature in triangular distribution
     real(rk) :: Emax  !-- max. excitation energy from Q-fission (MeV)
     real(rk) :: Eexc  !-- max. excitation energy for neutron emission (MeV)
     real(rk) :: T     !-- temperature for neutron emission
     integer  :: Nxs   !-- number of energy points on inverse x/s grid
     real(rk), dimension(:), pointer :: exs !-- energy grid for inverse x/s
     real(rk), dimension(:), pointer :: xs  !-- inverse x/s values
     real(rk) :: sigma0, sigma1 !-- values used to extrapolate x/s at low-E
     real(rk) :: E0, E1         !-- sigma0=sigma(E0) ; sigma1=sigma(E1)
  end type nucleus

  !---------------------------------------------------------------------------
  !-- ENERGY GRID representation
  !---------------------------------------------------------------------------
  type egrid
     integer :: n                              !-- number of points
     real(rk), dimension(:), pointer :: x  !-- points on the grid
     real(rk), dimension(:), pointer :: dx !-- mesh-size as a function of x
  end type egrid

  !---------------------------------------------------------------------------
  !-- SPECTRUM representation
  !---------------------------------------------------------------------------
  type spectrum
     real(rk), dimension(:), pointer :: spectrum
     real(rk), dimension(2) :: moments
  end type spectrum

  !------------------------ << DECLARATIONS >> -------------------------------

  type(pfnsType) :: pfns !-- [PUBLIC] the results!

  !-- declaration of compound nucleus, and light and heavy fragments
  type(nucleus), dimension(Nfc), target :: parentNucleus
  type(nucleus), dimension(Nfc), target :: lightFragment, heavyFragment

  !-- pointers to current parent nucleus and fragments
  type(nucleus), pointer :: parent, light, heavy

      !-- option for inverse compound formation cross-section
      !-- (constant or energy-dependent); default: constant
      character(len=10) :: csOption = "constant"

  !-- multiple-chance fission probabilities 
  real(rk), dimension(Nfc) :: Pf = 0.0_rk

  !-- actual number of multiple-chance fission probabilities (user input)
  integer :: numberChanceFissions

  !-- neutron binding energies (MeV)
  real(rk), dimension(Nfc) :: Bn

  real(rk) :: Einc !-- incident neutron energy (MeV)

  !-- level density parameters for the parent nucleus and the two fragments
  real(rk), dimension(Nfc) :: alev_p=0._rk
  real(rk), dimension(Nfc) :: alev_l=0._rk
  real(rk), dimension(Nfc) :: alev_h=0._rk

  !-- average fission-fragment total kinetic energy TKE (MeV) for the i^th-chance fission
  real(rk), dimension(Nfc) :: TKE

  !-- average energy release for each fission chance
  real(rk), dimension(Nfc) :: Erelease

  !-- average FF neutron separation energy
  real(rk), dimension(Nfc) :: Sn

  type(egrid) :: Eout !-- outgoing energy grid characteristics

  real(rk), dimension(NLeg) :: XLeg, WLeg !-- Gauss-Legendre abscissas & weights
  real(rk), dimension(NLag) :: XLag, WLag !-- Gauss-Laguerre abscissas & weights

  real(rk) :: E !-- energy (MeV)

  !-- average total prompt gamma-ray energy (MeV)
  real(rk), dimension(Nfc) :: Egamma = 0._rk

  real(rk), dimension(NLeg,NLag) :: interpolatedXSLight, interpolatedXSHeavy, &
       interpolatedXSParent

  !-- center-of-mass fission spectra for multiple-chance fission
  type(spectrum), dimension(Nfc) :: labLightSpectrum
  type(spectrum), dimension(Nfc) :: labHeavySpectrum
  type(spectrum), dimension(Nfc) :: labSpectrum

  !-- laboratory frame fission spectra for multiple-chance fission
  type(spectrum), dimension(Nfc) :: cmLightSpectrum
  type(spectrum), dimension(Nfc) :: cmHeavySpectrum
  type(spectrum), dimension(Nfc) :: cmSpectrum

  !-- center-of-mass prefission evaporation spectra from parent nuclei (A, A-1, ...)
  type(spectrum), dimension(Nfc) :: cmParentSpectrum

  !-- total laboratory frame fission spectrum (inc. all chance(s) fission)
  type(spectrum) :: labTotalSpectrum

  !-- neutron multiplicity as a function of i^th-chance fission
  real(rk), dimension(Nfc) :: nubar

  !-- total neutron multiplicity (inc. all chance(s) fission)
  real(rk) :: totalNubar
 
  !-- (sig0,sig1) parameters used to extrapolate cross sections at low energy
  real(rk) :: sigma0Light, sigma1Light
  real(rk) :: sigma0Heavy, sigma1Heavy
  real(rk) :: sigma0Parent, sigma1Parent

  real(rk) :: E0Light, E1Light, E0Heavy, E1Heavy
    
!-------
contains
!-------


  !---------------------------------------------------------------------------
  !                            SUBROUTINE pfnsCalc
  !---------------------------------------------------------------------------
  !
  ! Main subroutine.
  !
  ! First, it reads the main input file, and set the initial parameters 
  ! entering in the calculation.
  !
  ! It then computes Gauss-Legendre and Gauss-Laguerre abscissas and weights
  ! used in the following numerical integrations. The number of integration
  ! points are defined by the integer parameters NLeg and NLag, declared in 
  ! the header of the module.
  !
  ! The 'setup' subroutine is then called to initialize a set of variables
  ! needed in the calculation. In particular, it allocates the memory required
  ! by the spectra of the light and heavy fragments, and the average, in both
  ! the center-of-mass and laboratory reference frames. It also computes the
  ! maximum value of the triangular-shaped temperature distribution in both
  ! fragments.
  !
  ! The subroutine 'interpolationOnGaussPoints' performs the interpolation of the
  ! absorption cross sections for the light and heavy fragments, on the
  ! spatial defined defined by the Gauss-Legendre and Gauss-Laguerre points.
  ! The interpolated values are needed in the following calculation loops, 
  ! and their computation outside the loops helps reduce the CPU running time.
  !
  !---------------------------------------------------------------------------
  subroutine pfnsCalc

!    real(rk) :: x 

!    integer :: i

    !-- read main input file
    call readInput

    !-- set up various needed quantities
    call setup
        
    !-- main computational subroutine
    call computeSpectrum

    !-- write out results
    call resultsOut

    !-- create gnuplot files
    call resultsToGnuplot
  
    return
  end subroutine pfnsCalc


  !---------------------------------------------------------------------------
  !                     FUNCTION computeEnergyRelease
  !---------------------------------------------------------------------------
  ! Computes the energy release in the fission of the compound nucleus (Zc,Ac)
  ! into the light fragment (Zl,Al) and the heavy fragment (Zh,Ah).
  !
  ! Nuclear masses are taken from the RIPL-2 library.
  !
  !---------------------------------------------------------------------------
  real(rk) function computeEnergyRelease (zaid0, zaid1, zaid2)

    integer, intent(in) :: zaid0         !-- ZAID of fissioning nucleus
    integer, intent(in) :: zaid1, zaid2  !-- ZAID of light and heavy fragments

    integer :: Z0, A0
    integer :: Z1, A1, Z2, A2

    A0 = mod(zaid0,1000) ; Z0 = int(zaid0/1000)
    A1 = mod(zaid1,1000) ; Z1 = int(zaid1/1000)
    A2 = mod(zaid2,1000) ; Z2 = int(zaid2/1000)

    computeEnergyRelease = ripl2_masses(Z0,A0) - ripl2_masses(Z1,A1) &
         - ripl2_masses(Z2,A2)

    return
  end function computeEnergyRelease


  !---------------------------------------------------------------------------
  !                      SUBROUTINE computeSigma0Sigma1
  !---------------------------------------------------------------------------
  ! At sufficiently low neutron energy, the compound nucleus cross section is
  ! proportional to 1/v, where v is the neutron velocity. Following MN82, the
  ! optical model compound nucleus cross section is represented by
  !
  !        sig(E) = alpha + beta / sqrt(E)
  !
  ! The parameters alpha and beta are obtained by solving the above equation
  ! at 0.1 and 1.0 keV. See Equation (25) in [MN82].
  !---------------------------------------------------------------------------
  subroutine computeSigma0Sigma1 (cn)
  
    type(nucleus) :: cn
  
    real(rk) :: xs0, xs1
    real(rk) :: E0, E1, E0Sqr, E1Sqr
  
    !-------------------------------------------------------------------------
    !-- From D.G.Madland (on 04/28/2008),
    !--
    !-- sigma0 and sigma1 values to determine sigma at lower- and higher-
    !-- energies than the OMP energy grid read in input
    !--
    !-- sigma = sigma0 + sigma1 / sqrt (center-of-mass energy)
    !--
    !-- at the low energy end, at and below 1 keV, and sigma=sigma(37 MeV) for
    !-- center-of-mass energies of 37 MeV and above.
    !--
    !-- bg235l:     96-Sr (0.1 keV => 56.1004 b ; 1.0 keV => 17.8515 b)
    !--            140-Xe (0.1 keV => 125.6622 b ; 1.0 keV => 39.0939 b)
    !--
    !-- bg235star:  96-Sr same as above
    !--            139-Xe (0.1 keV => 124.2635 b ; 1.0 keV => 36.6706 b)
    !--
    !-- bg234star:  95-Sr (0.1 keV => 56.5513 b ; 1.0 keV => 17.9876 b)
    !--            139-Xe same as above
    !-------------------------------------------------------------------------
    
    E0=0.0001_rk !-- 0.1 keV
    E1=0.001_rk  !-- 1.0 keV

    select case (cn%zaid)
      case (38096)
        xs0 = 56.1004_rk !-- barns
        xs1 = 17.8515_rk
      case (38095)
        xs0 = 56.5513_rk
        xs1 = 17.9876_rk
      case (54140)
        xs0 = 125.6622_rk
        xs1 = 39.0939_rk
      case (54139)
        xs0 = 124.2635_rk
        xs1 = 38.6706_rk
      case default
        xs0 = cn%xs(1)*0.001 !-- at 1st energy point E0 (in barns)
        xs1 = cn%xs(2)*0.001 !-- at 2nd energy point E1 (in barns)
        E0 = cn%exs(1)
        E1 = cn%exs(2)
    end select
    
    E0Sqr = sqrt(E0)
    E1Sqr = sqrt(E1)
      
    !-- solve sig(E)=sigma0+sigma1/sqrt(E) equation for low-energy E
    !-- for both the light and heavy fragments
    cn%sigma0 = (E0Sqr*xs0-E1Sqr*xs1)/(E0Sqr-E1Sqr)
    cn%sigma1 = E0Sqr*E1Sqr/(E0Sqr-E1Sqr)*(xs1-xs0)

    cn%E0 = E0
    cn%E1 = E1

    if (cn%zaid==92236) then
      cn%sigma0 = 0.1900167662_rk
      cn%sigma1 = 0.4530958623_rk
      cn%E0 = 0.0001_rk
      cn%E1 = 0.001_rk
    endif
    
    if (cn%zaid==92235) then
      cn%sigma0 = 0.1849019088_rk
      cn%sigma1 = 0.4568654509_rk
      cn%E0 = 0.0001_rk
      cn%E1 = 0.001_rk
    endif

    return
  end subroutine computeSigma0Sigma1

  !---------------------------------------------------------------------------
  !                   SUBROUTINE interpolationOnGaussPoints
  !---------------------------------------------------------------------------
  ! Interpolate the inverse cross sections for the light and heavy fragments 
  ! and parent fissioning nucleus on the energy grid points specified by 
  ! Gauss-Legendre and Gauss-Laguerre abscissas.
  !---------------------------------------------------------------------------
  subroutine interpolationOnGaussPoints
    
    integer :: i,j
    real(rk) :: Tl, Th, Tp

    !-- compute interpolated cross sections at the quadrature points
    do i=1, NLeg
       Tl = light%Tmax*(1+XLeg(i))*0.5_rk
       Th = heavy%Tmax*(1+XLeg(i))*0.5_rk
       do j=1, NLag
          interpolatedXSLight(i,j) = interpolate(light,XLag(j)*Tl)
          interpolatedXSHeavy(i,j) = interpolate(heavy,XLag(j)*Th)
       end do
    end do

    !-- only one temperature in parent nucleus
    Tp = parent%T
    do j=1, NLag
      interpolatedXSParent(:,j) = interpolate (parent,XLag(j)*Tp)
    enddo
    
    return
  end subroutine interpolationOnGaussPoints


  !---------------------------------------------------------------------------
  !                    SUBROUTINE computeSpectrum
  !---------------------------------------------------------------------------
  ! Main computational subroutine.
  !---------------------------------------------------------------------------
  subroutine computeSpectrum

    integer :: ifc, j
    real(rk) :: x
  
    do ifc=1, numberChanceFissions  !-- loop over multiple-chance fissions
    
       parent => parentNucleus(ifc)
       light  => lightFragment(ifc)
       heavy  => heavyFragment(ifc)
       
       !-- calculate new Emax and Tmax for all nuclei in the i^th-chance fission
       if (ifc>=2) then

          !-- calculate evaporation c.m. spectrum
          if (csOption=="constant") then
            call computeParentSpectrumConstant (parentNucleus(ifc-1), &
              cmParentSpectrum(ifc-1))
          else
            call computeParentSpectrum (parentNucleus(ifc-1), &
              cmParentSpectrum(ifc-1), interpolatedXSParent)
          end if
            
          parent%Eexc = Einc - sum(cmParentSpectrum(1:ifc-1)%moments(1)) - sum(Bn(2:ifc))
          parent%T    = sqrt(parent%Eexc/parentNucleus(ifc-1)%alev)
          
          x=0._rk
          do j=1, ifc-1
             x=x+Bn(j)+cmParentSpectrum(j)%moments(1)
          end do

          parent%Emax = Erelease(ifc)+Einc+Bn(1)-TKE(ifc) - x
          parent%Tmax = sqrt(parent%Emax/parent%alev)

          light%Tmax = parent%Tmax
          heavy%Tmax = parent%Tmax

       end if

       if (csOption=="constant") then
       
          call computeCmSpectrumConstant (light, cmLightSpectrum(ifc))
          call computeCmSpectrumConstant (heavy, cmHeavySpectrum(ifc))
          
          call spectraAverage (cmLightSpectrum(ifc), cmHeavySpectrum(ifc), &
            cmSpectrum(ifc))
          
          call computeLabSpectrumConstant (light, labLightSpectrum(ifc))
          call computeLabSpectrumConstant (heavy, labHeavySpectrum(ifc))
          
          call spectraAverage (labLightSpectrum(ifc), labHeavySpectrum(ifc), &
            labSpectrum(ifc))
          
       else

          !-- prepare interpolated cross-sections on GL grids
          call interpolationOnGaussPoints
                    
          call computeCmSpectrum (light, cmLightSpectrum(ifc), &
            interpolatedXSLight)
          call computeCmSpectrum (heavy, cmHeavySpectrum(ifc), &
            interpolatedXSHeavy)
          
          call spectraAverage (cmLightSpectrum(ifc), cmHeavySpectrum(ifc), &
            cmSpectrum(ifc))

          call computeLabSpectrum (light, labLightSpectrum(ifc), &
            interpolatedXSLight)
          call computeLabSpectrum (heavy, labHeavySpectrum(ifc), &
            interpolatedXSHeavy)

          call spectraAverage (labLightSpectrum(ifc), labHeavySpectrum(ifc), &
            labSpectrum(ifc))
                              
        end if
                
        call multiplicity (ifc) !-- compute neutron multiplicity for ifc^th chance fission

    end do !-- end loop over multiple-chance fissions

    !-- computes total spectrum in laboratory frame (Eq.(53) in [MN82])
            call computeTotalSpectrum
    
    return
  end subroutine computeSpectrum


  !---------------------------------------------------------------------------
  !                    SUBROUTINE computeTotalSpectrum
  !---------------------------------------------------------------------------
  ! Computes total spectrum and multiplicity in laboratory frame.
      ! Eq.(53) in [MN82].
  !---------------------------------------------------------------------------
      subroutine computeTotalSpectrum
      
    real(rk) ::denominator !-- help variable
    real(rk), dimension(Eout%n) :: cmsum

    real(rk) :: dE

            integer  :: i, ifc
      
    allocate (labTotalSpectrum%spectrum(Eout%n))

    labTotalSpectrum%spectrum = Pf(1) * nubar(1) * labSpectrum(1)%spectrum
    denominator = Pf(1)*nubar(1)
    cmsum=0.
    do ifc=2, numberChanceFissions
       cmsum=cmsum+cmParentSpectrum(ifc-1)%spectrum
       labTotalSpectrum%spectrum = labTotalSpectrum%spectrum + &
            Pf(ifc) * ( cmsum + nubar(ifc)*labSpectrum(ifc)%spectrum )
       denominator = denominator + Pf(ifc)*(nubar(ifc)+ifc-1)
    end do

    labTotalSpectrum%spectrum = labTotalSpectrum%spectrum / denominator
    
    !-- computes 1st and 2nd moments of the laboratory total spectrum

    do i=1, Eout%n-1
       dE = Eout%dx(i)
       labTotalSpectrum%moments(1) = labTotalSpectrum%moments(1) + &
            Eout%x(i) * labTotalSpectrum%spectrum(i) * dE
       labTotalSpectrum%moments(2) = labTotalSpectrum%moments(2) + &
            Eout%x(i) * Eout%x(i) * labTotalSpectrum%spectrum(i) * dE
    end do

    !-- computes total neutron multiplicity (Eq.(57) in [MN82])
    totalNubar = Pf(1)*nubar(1)
    do ifc=2, numberChanceFissions
       totalnubar = totalnubar + Pf(ifc)*(nubar(ifc)+ifc-1)
    end do
    totalnubar = totalnubar / sum(Pf)

    return
  end subroutine computeTotalSpectrum
  
  
  !---------------------------------------------------------------------------
  !                 SUBROUTINE computeLabSpectrumConstant
  !---------------------------------------------------------------------------
  ! Computes the spectrum of outgoing neutron energy in the laboratory frame
  ! for the fission fragment FRAGMENT (intent(in)), in the case of a 
  ! CONSTANT compound formation CROSS-SECTION.
  ! This calculation is based on the formulas found in [MN'82], Appendix A, 
  ! which are closed form solutions.
  !---------------------------------------------------------------------------
  subroutine computeLabSpectrumConstant (fragment, fragmentSpectrum)
    
    type(nucleus), intent(in)   :: fragment
    type(spectrum), intent(out) :: fragmentSpectrum
      
    real(rk) :: u1, u2 !-- change of variables
    real(rk) :: Efsq   !-- sqrt(Ef) with Ef = fragment kinetic energy
    real(rk) :: Tm     !-- max. temperature in fragment    
    real(rk) :: coef   !-- help variable
    integer  :: i
    
    Efsq = sqrt(fragment%KE)
    Tm = fragment%Tmax
    
    coef = 1._rk/(3._rk*sqrt(fragment%KE*Tm))

    allocate (fragmentSpectrum%spectrum(Eout%n))

    do i=1, Eout%n !-- loop over outgoing energy grid
      E=Eout%x(i)
      u1=sqrt(E)-Efsq ; u1=u1*u1/Tm
      u2=sqrt(E)+Efsq ; u2=u2*u2/Tm
      fragmentSpectrum%spectrum(i) = coef * ( &
        u2**1.5*expInt(1,u2) - u1**1.5*expInt(1,u1) + &
        ( gammp(1.5_rk,u2) - gammp(1.5_rk,u1) ) * exp(gammln(1.5_rk)) )
    end do
    
    !-- eliminate very small negative values (~< 1e-18)
    where (fragmentSpectrum%spectrum<0.) fragmentSpectrum%spectrum=1e-18_rk
    
    fragmentSpectrum%moments(1) = firstMoment  (fragmentSpectrum%spectrum)
    fragmentSpectrum%moments(2) = secondMoment (fragmentSpectrum%spectrum)

    return
  end subroutine computeLabSpectrumConstant
      

  !---------------------------------------------------------------------------
  !                    SUBROUTINE computeLabSpectrum
  !---------------------------------------------------------------------------
  ! Computes the spectrum of outgoing neutron energy in the laboratory frame
  ! for the fission fragment FRAGMENT (intent(in)).
  !---------------------------------------------------------------------------
  subroutine computeLabSpectrum (fragment, fragmentSpectrum, xsint)

    type(nucleus), intent(in)   :: fragment
    type(spectrum), intent(out) :: fragmentSpectrum

    real(rk), dimension(NLeg,NLag), intent(in) :: xsint !-- interpolated x/s

    real(rk) :: I1, I2 !-- integration limits

    real(rk) :: X, Y, T, ELeg !-- help variables
    real(rk) :: Tmax, KE
    real(rk) :: N !-- normalization factor

    integer :: i, j, k

    allocate (fragmentSpectrum%spectrum(Eout%n))

    Tmax = fragment%Tmax
    KE   = fragment%KE

    do i=1, Eout%n-1 !-- loop over outgoing energy grid
       
       E=Eout%x(i)
       
       !-- limits of integration for fission fragment
       I1 = sqrt(E)-sqrt(KE) ; I1 = I1*I1
       I2 = sqrt(E)+sqrt(KE) ; I2 = I2*I2
  
       Y=0.
       do j=1, NLeg !-- transformation c.m. to lab
          
          ELeg = ( (I2-I1)*XLeg(j)+(I2+I1) )*0.5_rk
          
          X=0.
          !-- integration over temperature distribution
          do k=1, NLeg
            T = Tmax*(1+XLeg(k))*0.5_rk
            N = normalizationSpectrum (T, xsint(k,:))
            X = X + WLeg(k) * N * T * exp(-ELeg/T)
          end do
            
          Y = Y+WLeg(j)*sqrt(ELeg)*0.5_rk*Tmax*X*interpolate(fragment,ELeg)
                                    
       end do
       
       fragmentSpectrum%spectrum(i) = (I2-I1)*0.5/(2.*Tmax*Tmax*sqrt(KE))*Y
              
    end do

    !-- eliminate very small negative values (~< 1e-18)
    where (fragmentSpectrum%spectrum<0.) fragmentSpectrum%spectrum=1e-18_rk

    fragmentSpectrum%moments(1) = firstMoment  (fragmentSpectrum%spectrum)
    fragmentSpectrum%moments(2) = secondMoment (fragmentSpectrum%spectrum)
    
    return
  end subroutine computeLabSpectrum


  !---------------------------------------------------------------------------
  !                  SUBROUTINE computeCmSpectrumConstant
  !---------------------------------------------------------------------------
  ! Computes the spectrum of outgoing neutron energy in the center-of-mass
  ! frame for the fission fragment FRAGMENT (intent(in)), in the case of a
  ! CONSTANT compound formation CROSS-SECTION.
  !---------------------------------------------------------------------------
  subroutine computeCmSpectrumConstant (fragment, fragmentSpectrum)
      
    type(nucleus), intent(in)   :: fragment
    type(spectrum), intent(out) :: fragmentSpectrum
    
!   real(rk) :: E, dE
    real(rk) :: E
    real(rk) :: Tm     !-- max. Temperature for fragment
    real(rk) :: Tm2    !-- Tm**2
    integer :: i
    
    allocate (fragmentSpectrum%spectrum(Eout%n))
    fragmentSpectrum%moments=0._rk

    Tm  = fragment%Tmax
    Tm2 = Tm * Tm

    do i=1, Eout%n !-- loop over energy
      E=Eout%x(i)
      fragmentSpectrum%spectrum(i) = E*expInt(1,E/Tm)
    enddo    
    fragmentSpectrum%spectrum(:) = 2._rk*fragmentSpectrum%spectrum(:)/Tm2
    
    !-- eliminate very small negative values (~< 1e-18)
    where (fragmentSpectrum%spectrum<0.) fragmentSpectrum%spectrum=1e-18_rk
    
    fragmentSpectrum%moments(1)  = firstMoment (fragmentSpectrum%spectrum)
    fragmentSpectrum%moments(2)  = secondMoment(fragmentSpectrum%spectrum)

    return
  end subroutine computeCmSpectrumConstant

  !---------------------------------------------------------------------------
  !                    SUBROUTINE computeCmSpectrum
  !---------------------------------------------------------------------------
  ! Computes the spectrum of outgoing neutron energy in the center-of-mass
  ! frame for the fission fragment FRAGMENT (intent(in)).
  !---------------------------------------------------------------------------
  subroutine computeCmSpectrum (fragment, fragmentSpectrum, xsint)

    type(nucleus), intent(in)   :: fragment
    type(spectrum), intent(out) :: fragmentSpectrum

    real(rk), dimension(NLeg,NLag), intent(in) :: xsint !-- interpolated x/s

    real(rk) :: X, T, E !-- help variables
    real(rk) :: N       !-- normalization factor
    real(rk) :: Tmax
    integer  :: i,j

    allocate (fragmentSpectrum%spectrum(Eout%n))
    fragmentSpectrum%moments=0._rk

    Tmax = fragment%Tmax

    do i=1, Eout%n !-- loop over energy

      E = Eout%x(i)
      X=0._rk
      do j=1, NLeg !-- loop over temperature
        T = Tmax * (1+XLeg(j)) * 0.5_rk
        N = normalizationSpectrum (T, xsint(j,:))
        X = X + WLeg(j) * N * T * exp(-E/T)
      end do

      fragmentSpectrum%spectrum(i) = E*interpolate(fragment,E)/Tmax*X

    end do !-- end loop over outgoing energy

    !-- eliminate very small negative values (~< 1e-18)
    where (fragmentSpectrum%spectrum<0.) fragmentSpectrum%spectrum=1e-18_rk
    
    fragmentSpectrum%moments(1)  = firstMoment (fragmentSpectrum%spectrum)
    fragmentSpectrum%moments(2)  = secondMoment(fragmentSpectrum%spectrum)

    return
  end subroutine computeCmSpectrum


  !---------------------------------------------------------------------------
  !                    SUBROUTINE computeParentSpectrum
  !---------------------------------------------------------------------------
  ! Computes the spectrum of neutrons emitted from the parent nucleus in 
  ! the multiple-chance fission (pre-fission neutron evaporation).
  !---------------------------------------------------------------------------
  subroutine computeParentSpectrum (cn, cnSpectrum, xsint)

    type(nucleus), intent(in)   :: cn
    type(spectrum), intent(out) :: cnSpectrum

    real(rk), dimension(NLeg,NLag), intent(in) :: xsint !-- interpolated x/s

    real(rk) :: E
    real(rk) :: N !-- normalization factor
    integer  :: i

    allocate (cnSpectrum%spectrum(Eout%n))
    cnSpectrum%moments=0._rk
    
    !-- compute normalization factor k(T) in Eq.(22) [MN82]
    N = normalizationSpectrum (cn%T, xsint(1,:))
    
    do i=1, Eout%n !-- loop over outgoing energy
      E = Eout%x(i)
      cnSpectrum%spectrum(i) = N * E * interpolate(cn,E) * exp(-E/cn%T)
    end do

    !-- eliminate very small negative values (~< 1e-18)
    where (cnSpectrum%spectrum<0.) cnSpectrum%spectrum=1e-18_rk
    
    cnSpectrum%moments(1)  = firstMoment (cnSpectrum%spectrum)
    cnSpectrum%moments(2)  = secondMoment(cnSpectrum%spectrum)

    return
  end subroutine computeParentSpectrum

  !---------------------------------------------------------------------------
  !                 SUBROUTINE computeParentSpectrumConstant
  !---------------------------------------------------------------------------
  ! Computes the spectrum of neutrons emitted from the parent nucleus in 
  ! the multiple-chance fission (pre-fission neutron evaporation), assuming
  ! a constant inverse formation cross section.
  !---------------------------------------------------------------------------
  subroutine computeParentSpectrumConstant (cn, cnSpectrum)

    type(nucleus), intent(in)   :: cn
    type(spectrum), intent(out) :: cnSpectrum

    real(rk) :: E, dE  !-- help variables

    real(rk) :: Tm, Tm2 !-- max. temperature (and its square)

    integer :: i

    allocate (cnSpectrum%spectrum(Eout%n))
    cnSpectrum%moments=0._rk

    Tm = cn%T
    Tm2 = Tm*Tm

    do i=1, Eout%n-1 !-- loop over outgoing energy

       E = Eout%x(i)
       dE = Eout%dx(i)

       cnSpectrum%spectrum(i) = 2.*E/Tm2*expInt(1,E/Tm)

       cnSpectrum%moments(1) = cnSpectrum%moments(1) + &
            cnSpectrum%spectrum(i)*E*dE
       cnSpectrum%moments(2) = cnSpectrum%moments(2) + &
            cnSpectrum%spectrum(i)*E*E*dE

    end do !-- end loop over outgoing energy
    
    !-- eliminate very small negative values (~< 1e-18)
    where (cnSpectrum%spectrum<0.) cnSpectrum%spectrum=1e-18_rk

    return
  end subroutine computeParentSpectrumConstant


  !---------------------------------------------------------------------------
  !                        SUBROUTINE outgoingEnergyGrid
  !---------------------------------------------------------------------------
  ! Define a default outgoing energy grid, if user has not provided one in the
  ! main input file.
  !---------------------------------------------------------------------------
  subroutine outgoingEnergyGrid

    !-- default outgoing energy grid parameters
    integer  :: nen
    real(rk) :: Emin
    real(rk) :: Emax
  

    integer :: numberStepsPerDecade

    integer :: i,j,k

    real(rk) :: E   

!   real(rk) :: E, dE !-- energy step (log. interpolation)
!    nen = 90
!    Emin = 0.01
!    Emax = 30.0
!    dE = exp ( (log(Emax)-log(Emin)) / nen )
    
    if (.not.associated(Eout%x)) then

      !-- Madland's outgoing energy grid
      !-- energy bounds in the lab system: 10^-5 to 30 (or 50) MeV
      !-- 90 energy steps / decade
      nen = 581
      Emin = 1.e-5_rk
      Emax = 50._rk
      Eout%n = nen
      allocate (Eout%x(nen), Eout%dx(nen))
      numberStepsPerDecade = 90

      k=0
      do i=1, nen
        do j=1, numberStepsPerDecade
          k=k+1
          E = 10**(i-1)*Emin
          Eout%dx(k) = (10**i-10**(i-1))*Emin/real(numberStepsPerDecade)
          Eout%x(k) = E+real(j-1)*Eout%dx(k)
          if (k>580) exit
        enddo
        if (k>580) exit
      enddo
          
    end if

    return
  end subroutine outgoingEnergyGrid


  !---------------------------------------------------------------------------
  !                         SUBROUTINE spectraAverage
  !---------------------------------------------------------------------------
  ! Averages spectrum1 and spectrum2 into spectrumAve.
  ! For now, this is a simple average, and the spectra 1 and 2 must be on the
  ! same energy grid.
  !---------------------------------------------------------------------------
  subroutine spectraAverage (spectrum1, spectrum2, spectrumAve)

    type(spectrum), intent(in) :: spectrum1, spectrum2
    type(spectrum) :: spectrumAve

    integer :: n

    if (.not.associated(spectrum1%spectrum) .or. .not.associated(spectrum2%spectrum)) &
         stop 'spectraAverage: array pointer not associated'

    n = size(spectrum1%spectrum)

    if (n/=size(spectrum2%spectrum)) &
         stop 'spectraAverage: two spectra of differing sizes'

    if (associated(spectrumAve%spectrum)) then
       if (size(spectrumAve%spectrum)/=n) &
            stop 'spectraAverage: wrong size spectrumAve'
    else
       allocate(spectrumAve%spectrum(n))
    end if

    spectrumAve%spectrum = 0._rk
    spectrumAve%moments  = 0._rk

    spectrumAve%spectrum   = 0.5_rk*(spectrum1%spectrum+spectrum2%spectrum)
    spectrumAve%moments(1) = 0.5_rk*(spectrum1%moments(1)+spectrum2%moments(1))
    spectrumAve%moments(2) = 0.5_rk*(spectrum1%moments(2)+spectrum2%moments(2))

    return
  end subroutine spectraAverage

  !-------------------------------------------------------------------
  !                       function firstMoment
  !-------------------------------------------------------------------
  ! Computes the first moment of a given function f. It uses the
  ! Simpson's rule for the integration.
  !-------------------------------------------------------------------
  function firstMoment (f)
    real(rk), dimension(:) :: f
    real(rk) :: firstMoment
    integer :: i, n
    real(rk), parameter :: dx = 0.01_rk
    real(rk) :: s
    real(rk), dimension(:), allocatable :: x,y
    
    n=int(Eout%x(Eout%n)/dx)+1
    allocate (x(n), y(n))
    do i=1, n
      x(i) = (i-1)*dx
      y(i) = interpol (Eout%x, f, x(i)) * x(i)
    enddo

    call SimpsonIntegration(n,dx,y,s)
    firstMoment=s
    deallocate (x,y)

    return
  end function firstMoment


  !-------------------------------------------------------------------
  !                       function secondMoment
  !-------------------------------------------------------------------
  ! Computes the second moment of a given function f. It uses the
  ! Simpson's rule for the integration.
  !-------------------------------------------------------------------
  function secondMoment (f)
    real(rk), dimension(:) :: f
    real(rk) :: secondMoment
    integer :: i, n
    real(rk), parameter :: dx = 0.01_rk
    real(rk) :: s
    real(rk), dimension(:), allocatable :: x,y

    n=int(Eout%x(Eout%n)/dx)+1
    allocate (x(n), y(n))
    do i=1, n
      x(i) = (i-1)*dx
      y(i) = interpol (Eout%x, f, x(i)) * x(i) * x(i)
    enddo

    call SimpsonIntegration(n,dx,y,s)
    secondMoment=s
    deallocate (x,y)

    return
  end function secondMoment


  !---------------------------------------------------------------------------
  !                         SUBROUTINE multiplicity
  !---------------------------------------------------------------------------
  ! Computes the prompt neutron multiplicity nubar from Eq.(33) in [MN82], for
  ! the n^th chance fission.
  !---------------------------------------------------------------------------
  subroutine multiplicity (n)

    integer, intent(in) :: n !-- n^th chance fission
    real(rk) :: TXE, sum
    integer :: i

    !-- n^th-chance average total excitation energy
    sum=0._rk
    do i=1, n-1
       sum=sum+Bn(i)+cmParentSpectrum(i)%moments(1)
    end do
    TXE = Erelease(n) + Bn(1) + Einc - TKE(n) - sum

    !-- prompt gamma energy systematics from E.Fort, '90 IAEA Vienna Meeting
    !-- Egamma/=0 means it was set by the user in the input file 
    if (Egamma(n)==0.) Egamma(n) = 0.02772*parent%A+0.0891

    nubar(n) = (TXE - Egamma(n))/(Sn(n)+cmSpectrum(n)%moments(1))

    return
  end subroutine multiplicity
  

  !---------------------------------------------------------------------------
  !                     FUNCTION normalizationSpectrum
  !---------------------------------------------------------------------------
  ! Computes the normalization factor k(T) appearing in Eq.(24) of Madland and
  ! Nix paper, Nucl. Sci. Eng 81, pp.213-271 (1982).
  !---------------------------------------------------------------------------
  real(rk) function normalizationSpectrum (T, xsint)
    
    real(rk), intent(in)   :: T        !-- temperature of the spectrum
    real(rk), dimension(:) :: xsint    !-- interpolated cross section on GL grid
    real(rk)               :: x, y     !-- help variables
    integer                :: i

    y=0._rk
    do i=1, NLag
       x=Xlag(i)
       y = y + WLag(i) * x * xsint(i)
    end do
    y = y * T * T
    normalizationSpectrum = 1._rk/y

  end function normalizationSpectrum


  !---------------------------------------------------------------------------
  !                         SUBROUTINE readInput
  !---------------------------------------------------------------------------
  ! Reads input parameters from external file. They are given as a namelist.
  ! Comment lines can be added anywhere using the character '!' at the 
  ! beginning of the line. The order in which the input parameters are given
  ! does not matter. Finally, some input parameters need to be given 
  ! explicitly while some others are only optional.
  !
  ! The following parameters are MANDATORY:
  !
  ! ----------------------------------------------------------
  !       Description                 Example
  ! ----------------------------------------------------------
  !
  ! ZAID parent nucleus:              ZAIDparent = 92236
  ! ZAID fragment:                    ZAIDfragment = 38096
  ! incident neutron energy (MeV):    incidentEnergy = 0.53
  ! average TKE (MeV):                totalKineticEnergy = 170.40
  !
  ! The following parameters are OPTIONAL:
  !
  ! ----------------------------------------------------------
  !       Description                 Example
  ! ----------------------------------------------------------
  !
  ! average energy release (MeV):     energyRelease = 185.896
  ! level density parameters:         levelDensityParameterLF = 8.0
  ! 
  ! If these optional parameters are not present, here is what happens:
  !
  ! > Zf is deduced by assuming a correction (+-0.5) to the unchanged charge
  !   distribution.
  ! > the level density parameter for a given fragment is taken as A/11.
  ! > the average energy release is obtained from masses taken from the RIPL
  !   database.
  ! 
  !---------------------------------------------------------------------------
  subroutine readInput

    integer, parameter :: fu = 10 !-- file unit number

!   character(len=50) :: line, keyword, value !-- help variables
    logical :: there !-- boolean checking existence of input file
 !   integer :: eof   !-- end of file flag

    !-- outgoing energy grid parameters
    integer :: NE               !-- number of points
    real(rk) :: Emin, Emax, dE  !-- min, max, and bin-size values

    !-- charge and mass of one fragment
    integer :: Zf, Af

    !-- charge and mass of the compound nucleus
    integer :: Zc, Ac

    integer :: i

    integer  :: ZAIDparent, ZAIDfragment
    real(rk) :: incidentEnergy

    character(len=100) :: crossSectionType
    character(len=100) :: cmOutgoingEnergyGrid = ""

    real(rk), dimension(Nfc) :: levelDensityParentNucleus = 0._rk
    real(rk), dimension(Nfc) :: levelDensityLightFragment = 0._rk
    real(rk), dimension(Nfc) :: levelDensityHeavyFragment = 0._rk

    real(rk), dimension(Nfc) :: neutronBindingEnergy    = 0._rk
    real(rk), dimension(Nfc) :: neutronSeparationEnergy = 0._rk
    real(rk), dimension(Nfc) :: totalGammaRayEnergy     = 0._rk
    real(rk), dimension(Nfc) :: totalKineticEnergy      = 0._rk
    real(rk), dimension(Nfc) :: energyRelease           = 0._rk
    real(rk), dimension(Nfc) :: fissionProbability      = 0._rk
    
    namelist / inputs / ZAIDparent, ZAIDfragment, incidentEnergy, &
      crossSectionType, neutronBindingEnergy, neutronSeparationEnergy, &
      totalKineticEnergy, energyRelease, fissionProbability, &
      totalGammaRayEnergy, levelDensityParentNucleus, &
      levelDensityLightFragment, levelDensityHeavyFragment, cmOutgoingEnergyGrid

    !-- initializations
    Af=0; Zf=0

    inquire (file=inputFile, exist=there)
    if (.not.there) stop 'main input file not found!'

    open (fu, file=inputFile, status='old', action='read')
    read (fu,NML=inputs)
	close (fu)
!    read (fu,NML=inputs,end=10)
!10  close (fu)
    
    Zc = int(ZAIDparent/1000)
    Ac = mod(ZAIDparent,1000)

    Zf = int(ZAIDfragment/1000)
    Af = mod(ZAIDfragment,1000)

    !-- inverse compound formation cross section type
    !-- either 'constant' (no need to read any x/s file) or 'energy'-dependent
    !-- (will read x/s file later).
    csOption = crossSectionType

    Einc   = incidentEnergy

    Bn=0.; TKE=0. !-- (arrays)
    i=0
    do i=1, Nfc
      Pf(i)       = fissionProbability(i)
      if (Pf(i)==0.0) exit
      Bn(i)       = neutronBindingEnergy(i)
      Sn(i)       = neutronSeparationEnergy(i)
      TKE(i)      = totalKineticEnergy(i)
      Erelease(i) = energyRelease(i)
      Egamma(i)   = totalGammaRayEnergy(i)
      if (levelDensityParentNucleus(i)/=0.) alev_p(i) = levelDensityParentNucleus(i)
      if (levelDensityLightFragment(i)/=0.) alev_l(i) = levelDensityLightFragment(i)
      if (levelDensityHeavyFragment(i)/=0.) alev_h(i) = levelDensityHeavyFragment(i)
    enddo
    
    if (Pf(i)==0) then !-- actual number of chance-fissions
      numberChanceFissions=i-1
    else
      numberChanceFissions=i
    endif
    
    !-- outgoing energy grid (#bins, Emin, Emax)
    if (cmOutgoingEnergyGrid/="") then
      read(cmOutgoingEnergyGrid,*) NE, Emin, Emax
      !-- set up emission energy grid in log-scale
      if (Emin==0.) then
        write(*,*) 'warning: setting Emin to 0.01 MeV'
        Emin=0.01
      end if
      dE = exp( (log(Emax)-log(Emin))/NE )
      Eout%n=NE
      allocate (Eout%x(NE),Eout%dx(NE))
      Eout%x(1) = Emin
      do i=2, NE
        Eout%x(i)  = Eout%x(i-1)*dE
        Eout%dx(i) = Eout%x(i)-Eout%x(i-1)
      end do
      Eout%dx(1) = Eout%dx(2)
    endif

    !-- check that all MANDATORY fields are present, and set up a few derived
    !-- quantities.

    if (Zc==0.or.Ac==0) &
         stop '>> input :: (Z,A) of parent nucleus missing'
    if (Af==0) stop '>> input :: (Z,A) of fragment missing'

    !-- set pointers to 1st-chance fission nuclei
    parent => parentNucleus(1)
    light  => lightFragment(1)
    heavy  => heavyFragment(1)

    parent = createNucleus(1000*Zc+Ac, alev=alev_p(1), KE=TKE(1))

    if (Af<parent%A/2) then !-- Af corresponds to the light fragment
       if (Zf==0) Zf = int ( real(Zc*Af)/real(Ac)+0.5 + 0.5 )
       light = createNucleus (1000*Zf+Af, alev=alev_l(1))
       heavy = createNucleus (1000*(Zc-Zf)+Ac-Af, alev=alev_h(1))
    else !-- Af corresponds to the heavy fragment
       if (Zf==0) Zf = int ( real(Zc*Af)/real(Ac)-0.5 + 0.5 )
       heavy = createNucleus (1000*Zf+Af, alev=alev_l(1))
       light = createNucleus (1000*(Zc-Zf)+Zc-Af, alev=alev_h(1))
    end if
    
    return
  end subroutine readInput

  !---------------------------------------------------------------------------
  !                            SUBROUTINE readCrossSections
  !---------------------------------------------------------------------------
  ! Read the file XSFILE (see parameters above) and look for the inverse 
  ! compound formation cross section for the nucleus CN (input).
  !---------------------------------------------------------------------------
  subroutine readCrossSections (cn)

    type(nucleus), intent(inout) :: cn !-- nucleus of interest

!   character(len=50) :: line, keyword, value !-- help variables
    character(len=50) :: line

    logical :: there !-- boolean checking existence of input file
    integer :: eof   !-- end of file flag

    integer, parameter :: fu = 11

    integer :: Z, A  !-- charge and mass of nucleus read in file
    integer :: nen   !-- number of energies

    integer :: i

    inquire (file=xsfile, exist=there)
    if (.not.there) stop 'readCrossSections :: cross section file not found!'

    open (fu, file=xsfile, status='old', action='read')

    rewind fu

    eof=0
    do while (eof==0) !-- loop until end of file is reached

       read(fu,'(a)',iostat=eof) line
       if (eof/=0) exit !-- end of file

       if (line(1:1)=='#') then
          line = adjustl(line(2:))
          read(line,*) Z, A, nen

          if (Z==cn%Z.and.A==cn%A) then !-- start reading x/s

             read(fu,'(a)',iostat=eof) line !-- skip one blank line
             
             cn%Nxs = nen                
             allocate(cn%exs(nen), cn%xs(nen))
             read(fu,*) (cn%exs(i), cn%xs(i), i=1, nen)

             return !-- exit routine since nucleus has been found and x/s read

          end if

       end if

    end do

    if (eof/=0) then
       write(*,*) 'readCrossSections :: cross section not found for (',&
            trim(int2char(cn%z)), ',' , trim(int2char(cn%a)) , ')'
       stop
    end if

    close (fu)

  end subroutine readCrossSections


  !---------------------------------------------------------------------------
  !                            SUBROUTINE SETUP
  !---------------------------------------------------------------------------
  !
  ! It first defines the outgoing energy grid from the grid parameters, either
  ! given by the user or taken as defaults from the header of this module.
  ! 
  ! Masses from RIPL2 (FRDM option for now) are used to obtain the energy 
  ! release values in the case there is no user input for this quantity.
  !
  ! Then, the light and heavy fragments, as well as the parent nucleus, are 
  ! created for all multiple-chance fission components.
  !
  ! Gauss-Legendre and Gauss-Laguerre abscissas and weights are 
  ! calculated, and the relevant compound formation cross sections are
  ! interpolated on this GL-quadrature grid for easy integration.
  !
  !---------------------------------------------------------------------------
  subroutine setup

    integer :: zaid, zaidh, zaidl
    integer :: i

    !-- set outgoing energy grid (if not defined by the user)
    call outgoingEnergyGrid

    !-- computes energy releases in multiple-chance fission
    call ripl2_readMassesFRDM
!    call ripl2_readMassesHFB
    zaid=parent%zaid ; zaidl=light%zaid ; zaidh=heavy%zaid
    do i=1, Nfc
      if (Erelease(i)==0) Erelease(i) = computeEnergyRelease (zaid, zaidl, zaidh)
       zaid = zaid-1 ; zaidl=zaidl-1 !-- keep the heavy fragment fixed
    end do

    !-- average kinetic energies per nucleon of the light and heavy fragments
    light%KE = (real(heavy%A)/real(light%A))*(TKE(1)/real(parent%A))
    heavy%KE = (real(light%A)/real(heavy%A))*(TKE(1)/real(parent%A))

    !-- max. temperature in triangular distribution
    parent%Emax = Erelease(1) + Bn(1) + Einc - TKE(1)
    parent%Tmax = sqrt(parent%Emax/parent%alev)
    
    parent%Eexc = Einc
    parent%T    = sqrt(parent%Eexc/parent%alev)
    
    !-- max. temperature in triangular distribution for light and heavy fragments
    !-- Equal temperature assumtion
    light%Tmax = parent%Tmax
    heavy%Tmax = parent%Tmax
 
    !-- create and initialize all nuclei in the case of multiple-chance fission
    do i=2, numberChanceFissions

       zaid  = parentNucleus(i-1)%zaid-1
       
       if (i==2) then
        zaidl = lightFragment(i-1)%zaid
        zaidh = heavyFragment(i-1)%zaid-1
      else
        zaidl = lightFragment(i-1)%zaid-1
        zaidh = heavyFragment(i-1)%zaid
      endif

       parentNucleus(i) = createNucleus (zaid, KE=TKE(i), alev=alev_p(i))
       lightFragment(i) = createNucleus (zaidl, alev=alev_l(i))
       heavyFragment(i) = createNucleus (zaidh, alev=alev_h(i))

       lightFragment(i)%KE = (real(heavyFragment(i)%A)/real(lightFragment(i)%A)) * &
            (TKE(i)/real(parentNucleus(i)%A))
       heavyFragment(i)%KE = (real(lightFragment(i)%A)/real(heavyFragment(i)%A)) * &
            (TKE(i)/real(parentNucleus(i)%A))

    end do

    !-- compute Gauss-Legendre and Gauss-Laguerre abscissas and weights
    call GauLag (XLag, WLag, 0._rk) ; call GauLeg (XLeg, WLeg)
    
    return
  end subroutine setup


  !---------------------------------------------------------------------------
  !                            FUNCTION createNucleus
  !---------------------------------------------------------------------------
  ! This function returns an object of type(nucleus) from a ZAID identifier.
  ! Optional arguments include the level density parameter 'alev' as well as
  ! the kinetic energy for this nucleus.
  ! This routine also defines the mass of the nucleus (from RIPL2) and read
  ! the inverse compound formation cross section as needed in spectra
  ! calculations.
  !---------------------------------------------------------------------------
  function createNucleus (zaid, alev, KE) result (cn)

    integer, intent(in) :: zaid !-- ZAID number of nucleus to be created
    real(rk), intent(in), optional :: alev !-- level density parameter 'a'
    real(rk), intent(in), optional :: KE   !-- kinetic energy (MeV)

    type(nucleus) :: cn
    
    cn%zaid=zaid
    cn%z=int(zaid/1000)
    cn%a=mod(zaid,1000)

    if (present(alev).and.alev/=0.) then
       cn%alev=alev
    else
       cn%alev = cn%a/11. !-- Madland original value in [MN82]
    end if

    if (present(KE)) cn%KE=KE !-- kinetic energy

    cn%mass = ripl2_masses(cn%z,cn%a) !-- g.s. mass in MeV from RIPL2

!    cn%Bn = ? -- RIPL2 does not contain everything

    !-- initialize and read inverse compound formation x/s from file
    !-- and extrapolate at low-E using sig(E)=sig0+sig1/sqrt(E)
    if (csOption/="constant") then
      call readCrossSections (cn)
      call computeSigma0Sigma1 (cn)
    endif
    
    return
  end function createNucleus


  !---------------------------------------------------------------------------
  !                            SUBROUTINE deleteNucleus
  !---------------------------------------------------------------------------
  ! This routine simply frees the memory used each time 'createNucleus' is 
  ! called.
  !---------------------------------------------------------------------------
  subroutine deleteNucleus (cn)
    type(nucleus) :: cn

    if (associated(cn%exs)) then
       nullify (cn%exs)
       nullify (cn%xs)
    end if

  end subroutine deleteNucleus


  !---------------------------------------------------------------------------
  !                        SUBROUTINE resultsToGnuplot
  !---------------------------------------------------------------------------
  ! Save the results in a format readable by GNUPLOT. It also creates a 
  ! GNUPLOT file that will read those data.
  !---------------------------------------------------------------------------
  subroutine resultsToGnuplot

    character(len=*), parameter :: dataFile     = 'data.pfns'
    character(len=*), parameter :: multipleFile = 'multiple.pfns'
    character(len=*), parameter :: gnuplotFile  = 'pfns.plt'
    character(len=*), parameter :: psFile       = 'pfns.ps'

    integer, parameter :: fd = 10 !-- file unit for datafile
    integer, parameter :: fg = 11 !-- file unit for gnuplotfile

    integer :: i

    !-------------------------------
    !-- write results in data file -
    !-------------------------------
    open (fd, file=dataFile, action='write')

    write(fd,'(a)') '#'
    write(fd,'(a)') '# File created by the code PFNS_MN.F90 Version beta'
    write(fd,'(a)') '#'
    write(fd,'(a)') '# Author: P.Talou (talou@lanl.gov), LANL, USA'
    write(fd,'(a)') '#'
    write(fd,'(a)') '# Prompt Fission Neutrons Spectrum'
    write(fd,'(a)') '#'
    write(fd,'(a,a2,a1,a3,a,f5.2,a)') &
        '# n + (', trim(int2char(parentNucleus(1)%z)), ',', &
        trim(int2char(parentNucleus(1)%a-1)), ') at En = ', Einc, ' MeV'


    write(fd,'(a)') '#'
    write(fd,'(a)') '#     En''     lab. light   lab. heavy    lab. ave.  lab. Total'
    write(fd,'(a)') '#    (MeV)       (1/MeV)      (1/MeV)      (1/MeV)     (1/MeV) '
    write(fd,'(a)') '#--------------------------------------------------------------'

    do i=1, Eout%n !-- loop over energies on outgoing grid
       write(fd,'(1p,5(e12.5,1x))') Eout%x(i), labLightSpectrum(1)%spectrum(i), &
            labHeavySpectrum(1)%spectrum(i), labSpectrum(1)%spectrum(i), &
            labTotalSpectrum%spectrum(i)
    end do

    close (fd)

    !----------------------------------
    ! If multiple-chance fission only -
    !----------------------------------
    if (numberChanceFissions>1) then
    
      open (fd, file=multipleFile, action='write')

      write(fd,'(a)') '#'
      write(fd,'(a)') '# File created by the code PFNS_MN.F90 Version beta'
      write(fd,'(a)') '#'
      write(fd,'(a)') '# Author: P.Talou (talou@lanl.gov), LANL, USA'
      write(fd,'(a)') '#'
      write(fd,'(a)') '# Prompt Fission Neutrons Spectrum'
      write(fd,'(a)') '#'
      write(fd,'(a,a2,a1,a3,a,f5.2,a)') &
        '# n + (', trim(int2char(parentNucleus(1)%z)), ',', &
        trim(int2char(parentNucleus(1)%a-1)), ') at En = ', Einc, ' MeV'

      if (numberChanceFissions==2) then
      
        write(fd,'(a)') '#'
        write(fd,'(a)') '#     En''      1st-chance   2nd-chance   lab. Total'
        write(fd,'(a)') '#    (MeV)      (1/MeV)      (1/MeV)       (1/MeV)  '
        write(fd,'(a)') '#---------------------------------------------------'

        do i=1, Eout%n !-- loop over energies on outgoing grid
          write(fd,'(1p,5(e12.5,1x))') Eout%x(i), labSpectrum(1)%spectrum(i), &
            labSpectrum(2)%spectrum(i), labTotalSpectrum%spectrum(i)
        end do
      
      elseif (numberChanceFissions>=3) then

        write(fd,'(a)') '#'
        write(fd,'(a)') '#     En''      1st-chance   2nd-chance   3rd-chance   lab. Total'
        write(fd,'(a)') '#    (MeV)      (1/MeV)      (1/MeV)       (1/MeV)      (1/MeV)  '
        write(fd,'(a)') '#----------------------------------------------------------------'

        do i=1, Eout%n !-- loop over energies on outgoing grid
          write(fd,'(1p,5(e12.5,1x))') Eout%x(i), labSpectrum(1)%spectrum(i), &
            labSpectrum(2)%spectrum(i), labSpectrum(3)%spectrum(i), &
            labTotalSpectrum%spectrum(i)
        end do

      end if

      
      close (fd)
    
    end if

    !-----------------------
    !-- build gnuplot file -
    !-----------------------
    open (fg, file=gnuplotFile, action='write')

    write(fg,'(a,/)') 'reset'

    write(fg,'(a)') 'set term postscript enhanced 20 color'
    write(fg,'(a,/)') 'set output "' // trim(psFile) // '"'

    write(fg,'(a)') 'set log y'
    write(fg,'(a)') 'set mytics 10'
    write(fg,'(a,/)') 'set mxtics 5'

    write(fg,'(a)') 'set xrange [0:15]'
    write(fg,'(a,/)') 'set yrange [1e-5:1]'

    write(fg,'(a)') 'set xlabel "Laboratory neutron energy (MeV)"'
    write(fg,'(a,/)') 'set ylabel "Neutron energy spectrum N(E) (1/MeV)"'

    write(fg,'(a,/)') 'set key left bottom' 

!    write(fg,'(a,/)') 'set title "n+(' // trim(int2char(parent%z)) // ',' &
!         // trim(int2char(parent%a)) // ') at En = ' // Einc // ' MeV"'

    write(fg,'(a)') 'plot "' // trim(dataFile) // '" u 1:2 w l ti ' &
         // trim('"1st-chance lab. LF",\ ')
    write(fg,'(a)') '     "' // trim(dataFile) // '" u 1:3 w l ti ' &
         // trim('"1st-chance lab. HF",\ ')
    write(fg,'(a)') '     "' // trim(dataFile) // '" u 1:4 w l ti ' &
         // trim('"1st-chance lab. average",\ ')
    write(fg,'(a)') '     "' // trim(dataFile) // &
         '" u 1:5 w l ti "multiple-chance lab. average"'

    close (fg)

    return
  end subroutine resultsToGnuplot



  !---------------------------------------------------------------------------
  !                           SUBROUTINE resultsOut
  !---------------------------------------------------------------------------
  ! Write out some results.
  !---------------------------------------------------------------------------
  subroutine resultsOut

    character(3), dimension(5) :: nth = (/'1st','2nd','3rd','4th','5th'/)

    integer :: i, nen

    integer, parameter :: iout = 60 !-- unit file number for main output

    !-------------------------------------------------------------------------
    !-- write results in pfnsSpectrum (which is PUBLIC)
    !-------------------------------------------------------------------------

    nen=Eout%n

    pfns%numEout          = nen
    pfns%numChanceFission = numberChanceFissions

    allocate (pfns%Eout(nen))
    allocate (pfns%labTotalSpectrum(nen))
    allocate (pfns%labChanceSpectrum(numberChanceFissions,nen))
    allocate (pfns%cmChanceSpectrum(numberChanceFissions,nen))
    allocate (pfns%aveChanceEcm(numberChanceFissions))
    allocate (pfns%aveChanceElab(numberChanceFissions))
    allocate (pfns%chanceNubar(numberChanceFissions))

    pfns%Eout = Eout%x
!    pfns%cmTotalSpectrum
    pfns%labTotalSpectrum = labTotalSpectrum%spectrum

    do i=1, Nfc !-- loop over multiple-chance fission
       if (Pf(i)==0) exit
       pfns%labChanceSpectrum(i,:) = labSpectrum(i)%spectrum(:)
       pfns%cmChanceSpectrum(i,:)  = cmSpectrum(i)%spectrum(:)
       pfns%aveChanceEcm(i)        = cmSpectrum(i)%moments(1)
       pfns%aveChanceElab(i)       = labSpectrum(i)%moments(1)
       pfns%chanceNubar(i)         = nubar(i)
    end do

    pfns%aveElab    = labTotalSpectrum%moments(1)
    pfns%totalNubar = totalNubar

    !-------------------------------------------------------------------------

    parent => parentNucleus(1)
    light  => lightFragment(1)
    heavy  => heavyFragment(1)

    !-- open main output file
    open (iout, file=outputFile, action='write')

    write(iout,'(/,a50,/)') '--- Fission Neutron Spectrum Calculation ---'

    write(iout,'(a40,f5.2,/)') 'Incident neutron energy (MeV) : ', Einc

    write(iout,'(a40,i3,2x,i3)') '(Z,A) compound nucleus : ', parent%z, parent%a
    write(iout,'(a40,i3,2x,i3)') '(Z,A) light fragment   : ', light%z, light%a
    write(iout,'(a40,i3,2x,i3)') '(Z,A) heavy fragment   : ', heavy%z, heavy%a
    
    write(iout,'(/,a40,5f8.2)') 'Energy release(s) (MeV) : ', &
      Erelease(1:numberChanceFissions)
    write(iout,'(a40,5f8.2)') 'Average TKE(s) (MeV) : ', &
      TKE(1:numberChanceFissions)
    

    do i=1, numberChanceFissions !-- loop over multiple-chance fissions

       parent => parentNucleus(i)
       light  => lightFragment(i)
       heavy  => heavyFragment(i)

       write(iout,'(/,a20,a3,a20,/)') '*** ', nth(i), '-chance fission *** '

       write(iout,'(a10,a29,f5.3)') nth(i), '-chance fission probability: ', Pf(i)

       write(iout,'(/,a25,i3,a1,i3,a10,i3,a1,i3,a10,i3,a1,i3,a1,/)') &
            'Fissioning nucleus: (', parent%Z, ',', parent%A, &
            ') -- LF: (', light%Z, ',', light%A, ') -- HF: (', heavy%Z, ',', &
            heavy%A, ')'

       write(iout,'(a40,f7.3)') 'Max. average TXE (MeV): ', parent%Emax
       write(iout,'(a37,i1,a2,f7.3)') 'Max. temperature Tm', i, ': ', parent%Tmax
       write(iout,'(a37,i1,a2,f7.3)') 'Parent temperature T', i, ': ', parent%T

       write(iout,'(/,a40,f7.3,a10,f7.3,/)') &
            'Kinetic Energy per nucleon (MeV): [LF]: ', light%KE,' ; [HF]: ', heavy%KE
            
       write(iout,'(a40,f7.3,a10,f7.3,a15,f7.3)') &
            '<Ecm> (MeV): [LF]: ', cmLightSpectrum(i)%moments(1), ' ; [HF]: ', &
            cmHeavySpectrum(i)%moments(1), ' ; [Average]: ', cmSpectrum(i)%moments(1)
       write(iout,'(a40,f7.3,a10,f7.3,a15,f7.3)') &
            '<Ecm^2> (MeV): [LF]: ', cmLightSpectrum(i)%moments(2), ' ; [HF]: ', &
            cmHeavySpectrum(i)%moments(2), ' ; [Average]: ', cmSpectrum(i)%moments(2)
       write(iout,'(a40,f7.3,a10,f7.3,a15,f7.3)') &
            '<Elab> (MeV): [LF]: ', labLightSpectrum(i)%moments(1), ' ; [HF]: ', &
            labHeavySpectrum(i)%moments(1), ' ; [Average]: ', labSpectrum(i)%moments(1)
       write(iout,'(a40,f7.3,a10,f7.3,a15,f7.3,/)') &
            '<Elab^2> (MeV): [LF]: ', labLightSpectrum(i)%moments(2), ' ; [HF]: ', &
            labHeavySpectrum(i)%moments(2), ' ; [Average]: ', labSpectrum(i)%moments(2)
       write(iout,'(a40,f7.3)') &
            '<Ecm> parent (MeV): ', cmParentSpectrum(i)%moments(1)
       write(iout,'(a40,f7.3,/)') &
            '<Ecm^2> parent (MeV): ', cmParentSpectrum(i)%moments(2)
       write(iout,'(a40,f7.3,/)') 'multiplicity: ', nubar(i)
       write(iout,'(a40,f7.3)')   '<E^tot_gamma> (MeV) : ', Egamma(i)

    end do
    
    write(iout,'(/,a40,/)') '*** Total fission ***'
        
    write(iout,'(/,a40,f7.3)') '<Elab> (MeV): ', labTotalSpectrum%moments(1)
    write(iout,'(a40,f7.3)') '<Elab^2> (MeV): ', labTotalSpectrum%moments(2)
    
    write(iout,'(/,a40,f7.3)')   'Total multiplicity: ', totalNubar
        
    !    end do !-- end loop over multiple-chance fission
    
    
  end subroutine resultsOut


  !---------------------------------------------------------------------------
  !                            SUBROUTINE GauLag
  !---------------------------------------------------------------------------
  ! Given alf, the parameter alpha of the Laguerre polynomials, this routine
  ! returns arrays x[1..n] and w[1..n] containing the abscissas and weights
  ! of the n-point Gauss-Laguerre quadrature formula. The smallest abscissa
  ! is returned in x[1], the largest in x[n].
  !---------------------------------------------------------------------------
  ! Taken from Numerical Recipes, 2nd Ed., W.H.Press et al.
  !---------------------------------------------------------------------------
  subroutine GauLag (x, w, alf)
    real(rk), dimension(:) :: x, w
    real(rk) :: alf
    integer :: n
    integer :: i, its, j
    real(rk) :: ai, xn
    real(rk) :: p1, p2, p3, pp, z, z1

    integer :: MAXIT
    real(rk) :: EPS
    parameter (MAXIT=10, EPS=3._rk-14)

    n=size(x)
    if (size(w)/=n) stop 'array size mismatch !'

    do i=1,n
       if (i==1) then
          z=(1._rk+alf)*(3._rk+0.92_rk*alf)/(1._rk+2.4_rk*n+1.8_rk*alf)
       else if (i==2) then
          z=z+(15._rk+6.25_rk*alf)/(1._rk+0.9_rk*alf+2.5_rk*n)
       else
          ai=i-2
          z=z+((1._rk+2.55_rk*ai)/(1.9_rk*ai)+1.26_rk*ai*alf/ &
               (1._rk+3.5_rk*ai))*(z-x(i-2))/(1._rk+0.3_rk*alf)
       end if

       do its=1, MAXIT
          p1=1._rk
          p2=0._rk
          do j=1, n
             p3=p2
             p2=p1
             p1=((2._rk*j-1+alf-z)*p2-(j-1+alf)*p3)/real(j)
          end do
          pp=(n*p1-(n+alf)*p2)/z
          z1=z
          z=z1-p1/pp
          if (abs(z-z1)<=EPS) cycle
       end do
       !IF (its>MAXIT) WRITE(*,*) 'Too many iterations in GAULAG'
       x(i)=z
       xn = real(n)
       w(i)=-exp(Gammln(alf+n)-Gammln(xn))/(pp*n*p2)
    end do

  end subroutine GauLag


  !---------------------------------------------------------------------------
  !                              FUNCTION Gammln
  !---------------------------------------------------------------------------
  ! Returns the value ln(Gamma(xx))
  !---------------------------------------------------------------------------
  ! Numerical Recipes in C, 2nd Ed., p.214
  ! translated in F90 by P.Talou, 
  ! talou@lanl.gov
  !---------------------------------------------------------------------------
  real(rk) function Gammln (xx)
    real(rk), intent(in) :: xx
    real(rk) :: x, y, tmp, ser
    real(rk), dimension(6) :: cof = (/76.18009172947146_rk, &
         -86.50532032941677_rk, 24.01409824083091_rk, -1.231739572450155_rk, &
         0.1208650973866179e-2_rk, -0.5395239384953e-5_rk /)
    integer :: j

    x=xx
    y=x
    tmp=x+5.5_rk
    tmp = (x+0.5_rk)*log(tmp)-tmp
    ser = 1.000000000190015_rk
    do j=1, 6
       y=y+1._rk
       ser = ser + cof(j)/y
    end do
    Gammln = tmp+log(2.5066282746310005_rk*ser/x)
    return

  end function Gammln


  !---------------------------------------------------------------------------
  !                           FUNCTION gammp
  !---------------------------------------------------------------------------
  ! Returns the incomplete gamma function P(a,x):
  !
  ! P(a,x) = gamma(a,x)/Gamma(a) = 1/Gamma(a) int_0^x[exp(-t)t^(a-1)dt] (a>0)
  !
  !----------------------------------------------------------------------------
  ! Taken from "Numerical Recipes in Fortran 90".
  ! Uses gcf and gser
  !----------------------------------------------------------------------------
  real(rk) function gammp (a,x)
    real(rk), intent(in) :: a,x     
    real(rk) :: gammcf, gamser, gln
    if (x<0..or.a<=0.) stop 'bad arguments in gammp'
    if (x<a+1.) then !-- use the series representation
       call gser (gamser,a,x,gln)
       gammp=gamser
    else !-- use the continued fraction representation
       call gcf(gammcf,a,x,gln)
       gammp=1._rk-gammcf  !-- and take its complement.
    end if
    return  
  end function gammp
  
  !---------------------------------------------------------------------------
  !                           SUBROUTINE gser
  !---------------------------------------------------------------------------
  ! Returns the incomplete gamma function P(a,x) evaluated by its series
  ! representation as gamser.
  !----------------------------------------------------------------------------
  ! Taken from "Numerical Recipes in Fortran 90".
  !----------------------------------------------------------------------------
  subroutine gser (gamser,a,x,gln)

    real(rk), intent(in)  :: a,x
    real(rk), intent(out) :: gamser, gln
    
    integer, parameter :: itmax=100
    real(rk), parameter :: eps=3.e-7_rk
    
    integer  :: n
    real(rk) :: ap,del,sum
    
    gln=gammln(a)
    if (x<=0.) then
       if (x<0.) stop 'x<0 in gser'
       gamser=0._rk
       return
    end if
    ap=a
    sum=1._rk/a
    del=sum
    do n=1,itmax
       ap=ap+1._rk
       del=del*x/ap
       sum=sum+del
       if (abs(del)<abs(sum)*eps) goto 1
    enddo
    stop 'a too large, itmax too small in gser'
1   gamser=sum*exp(-x+a*log(x)-gln)
    
    return
  end subroutine gser


  !---------------------------------------------------------------------------
  !                           SUBROUTINE gcf
  !---------------------------------------------------------------------------
  ! Returns the incomplete gamma function Q(a,x) evaluated by its continued
  ! fraction representation as gammcf. Also returns ln(Gamma(a)) as gln.
  !----------------------------------------------------------------------------
  ! Taken from "Numerical Recipes in Fortran 90".
  !----------------------------------------------------------------------------
  subroutine gcf (gammcf,a,x,gln)
      
    real(rk), intent(in)  :: a,x
    real(rk), intent(out) :: gammcf,gln
    
    integer, parameter  :: itmax=100
    real(rk), parameter :: eps=3.e-7_rk
    real(rk), parameter :: fpmin=1.e-30_rk
    
    integer  :: i
    real(rk) :: an,b,c,d,del,h
    
    gln=gammln(a)
    b=x+1._rk-a                     !-- set up for evaluating continued fraction by modified
    c=1._rk/fpmin             !-- Lentz's method with b_0=0
    d=1._rk/b
    h=d
    do i=1,itmax
       an=-i*(i-a)
       b=b+2._rk
       d=an*d+b
       if (abs(d)<fpmin) d=fpmin
       c=b+an/c
       if (abs(c)<fpmin) c=fpmin
       d=1._rk/d
       del=d*c
       h=h*del
       if (abs(del-1._rk)<eps) goto 1
    end do
    stop 'a too large, itmax too small in gcf'
1   gammcf=exp(-x+a*log(x)-gln)*h
    
    return
  end subroutine gcf


  !---------------------------------------------------------------------------
  !                            SUBROUTINE GauLeg
  !---------------------------------------------------------------------------
  ! This routine returns arrays x[1..n] and w[1..n] containing
  ! the abscissas and weights of the n-point Gauss-Legendre
  ! quadrature formula. The smallest abscissa is returned in 
  ! x[1], the largest in x[n].
  !----------------------------------------------------------------------------
  ! Taken from Numerical Recipes, 2nd Ed., W.H.Press et al.
  !----------------------------------------------------------------------------
  subroutine GauLeg(X, W)

    real(rk), dimension(:) :: X, W
    integer :: N

    real(rk) :: EPS
    integer :: M

    integer :: I,J
    real(rk) :: Z,P1, P2,P3,PP,Z1
    parameter (EPS=3._rk*10._rk**(-14))

    real(kind=rk), parameter :: pi    = 3.14159265359_rk

    N=SIZE(X)
    if (SIZE(W)/=N) STOP 'array size mismatch !'

    M=(N+1)/2

    do I=1,M
       Z=cos(pi*(I-.25_rk)/(N+.5_rk))
1      continue
       P1=1._rk
       P2=0._rk
       do J=1,N
          P3=P2
          P2=P1
          P1=((2._rk*J-1._rk)*Z*P2-(J-1._rk)*P3)/J
       end do
       PP=N*(Z*P1-P2)/(Z*Z-1._rk)
       Z1=Z
       Z=Z1-P1/PP
       if (abs(Z-Z1).gt.EPS) go to 1

       X(I)=-Z
       X(N+1-I)=Z
       W(I)=2._rk/((1._rk-Z*Z)*PP*PP)
       W(N+1-I)=W(I)

    end do

  end subroutine GauLeg


  !---------------------------------------------------------------------------
  !                          FUNCTION interpolate
  !---------------------------------------------------------------------------
  ! Interpolates the cross section cn%xs at the energy x0.
  ! At low energy, uses MN82 approximation sig(E)=sig0+sig1/sqrt(E) to
  ! extrapolate the x/s.
  ! At higher energies, use the 'interpol' function.
  ! At energies above max. energy of the x/s energy grid, use a constant x/s.
  !---------------------------------------------------------------------------
  function interpolate (cn, x0)
  
    real(rk) :: interpolate !-- [result]
    real(rk), intent(in)   :: x0 !-- energy point at which to interpolate
    type(nucleus) :: cn !-- nucleus for which interpolation is calculated

    if (x0<cn%E1) then !-- low-energy treatment: sig(E)=sig0+sig1/sqrt(E)
      if(x0>0) then     ! RCN, Oct. 2009
        interpolate = cn%sigma0+cn%sigma1/sqrt(x0)
      else
        interpolate = cn%sigma0
      endif
      interpolate = interpolate*1000._rk
    elseif (x0>cn%exs(cn%Nxs)) then
      interpolate = cn%xs(cn%Nxs)      !-- constant x/s above Emax
    else
      interpolate = interpol (cn%exs, cn%xs, x0)
    endif
    
    return
  end function interpolate
  
    
  !---------------------------------------------------------------------------
  !                            FUNCTION interpol
  !---------------------------------------------------------------------------
  ! This function makes use of the GNASH subroutines INTERP and
  ! SPLINE in order to get an interpolated value of a given
  ! function y at a given value x=x_0: y(x=x_0).
  !---------------------------------------------------------------------------
  function interpol (x, y, x0)

    real(rk) :: interpol !-- result of the interpolation

    real(rk), dimension(:), intent(in) :: x !-- abscissa grid
    real(rk), dimension(:), intent(in) :: y !-- function to be interpolated

    real(rk), dimension(size(x)) :: f !-- spline of y

    real(rk), dimension(size(x)) :: xx
    real(rk), dimension(size(y)) :: yy

    real(rk), intent(in) :: x0 !-- x-value on which the function is interpolated

    real(rk) :: xx0

    integer :: n !-- size of the (x,y) arrays

    logical :: loginterp !-- turn on interpolation on logarithm scales

!    real(rk) :: res !-- result == what 'interpol' returns!

    !-- local variables
    real(rk), dimension(size(x)+3) :: dm
    integer :: i, nm1

    real(rk) :: dd1, dd2

    !-- local variables
    integer :: il, il1    !-- grid indexes
    real(rk) :: h, h2, h3    !-- powers of x-mesh
    real(rk) :: dx, dxmh
    real(rk) :: fh1, fh2, fh3, fh4
    real(rk) :: t1, t2

    n=size(x)
    if (size(y)/=n) stop &
         "size mismatch between x- and y-arrays in interpolation process"

    if (n.le.1) return

    loginterp = .false. !-- logarithm interpolation turned off by default

10  continue

    !-- to be able to modify x,y, and x0 if needed.
    xx=x ; yy=y ; xx0=x0

    !-- if logarithm interpolation, then modify arrays accordingly
    if (loginterp) then
       where (y/=0.)
          yy=log(y)
       elsewhere
          yy=-10 !-- i.e., y=1e-10
       end where
    end if

    nm1=n-1
    do i=1,nm1
       if (xx(i+1)==xx(i)) write(*,*) 'interpol ', i, i+1, xx(i), xx(i+1)
      dm(i+2)=(yy(i+1)-yy(i))/(xx(i+1)-xx(i))
    end do
    dm(2)=2.*dm(3)-dm(4)
    dm(1)=2.*dm(2)-dm(3)
    dm(n+2)=2.*dm(n+1)-dm(n)
    dm(n+3)=2.*dm(n+2)-dm(n+1)
    
    !-- function FUNDAK from [GNASH] inline here.

    do i=1,n
       dd1=dm(i+1)-dm(i); dd2=dm(i+3)-dm(i+2)
       if (dd1==0.or.dd2==0) then
          f(i)=0.5*(dm(i+1)+dm(i+2))
       else
          f(i)=(abs(dd2)*dm(i+1)+abs(dd1)*dm(i+2))/(abs(dd2)+abs(dd1))
       end if
    end do

    !-- function INTERP inline here.

    !-- test if xval outside grid boundaries
    if (n.le.1.or.xx0.lt.xx(1).or.xx0.gt.xx(n)) then
       interpol=0
       return
    end if

    !-- find 'il' such that x(il) < xval < x(il+1)
    do i = 2, n
       if (xx0 .le. xx(i)) exit
    end do
    il = i - 1                                                        
    il1=il+1                                                          

    !-- test if y(il)=y(il+1)
    if (yy(il)==yy(il1)) then
       interpol = yy(il)
       return
    end if

    !-- Dx, Dx*Dx, Dx*Dx*Dx
    h=xx(il1)-xx(il)
    h2=h**2
    h3=h**3

    dx=xx0-xx(il)
    dxmh=dx-h
    fh1=dxmh*(dx*(2.*dx-h)-h2)
    fh2=dx*dx*(2.*(-dxmh)+h)
    fh3=dx*dxmh*dxmh
    fh4=dx*dx*dxmh

    t1 = (yy(il)*fh1+y(il1)*fh2)/h3
    t2 = (f(il)*fh3+f(il1)*fh4)/h2
    interpol=t1+t2

    if (loginterp) then
       if (interpol>yy(il1)) interpol=yy(il1)
       if (interpol<yy(il))  interpol=yy(il)
       interpol = exp(interpol)
    end if

    if (interpol<0.) then
       loginterp=.true.
       goto 10
    end if

    return
  end function interpol


  !-------------------------------------------------------------------
  !                      FUNCTION expInt
  !-------------------------------------------------------------------
  ! Evaluates the exponential integral E_n(x).
  ! (based on Numerical Recipes)
  !-------------------------------------------------------------------
  real(rk) function expInt (n, x)

    integer, intent(in)  :: n
    real(rk), intent(in) :: x

    integer, parameter  :: maxIterations = 100     !-- max. allowed number of iterations
    real(rk), parameter :: epsilon = 1.0e-7_rk     !-- desired relative error
    real(rk), parameter :: FPmin = 1.0e-30_rk      !-- number near smallest representable floating-point number
    real(rk), parameter :: Euler = 0.5772156649_rk !-- Euler's constant (gamma)
    
    integer  :: i,ii,nm1
    real(rk) :: a,b,c,d,del,fact,h,psi

    nm1=n-1
    if (n<0.or.x<0..or.(x==0..and.(n==0.or.n==1))) then
       stop 'bad arguments in expInt'
    else if (n==0) then  !-- special case
       expInt=exp(-x)/x
    else if (x==0.) then !-- another special case
       expInt=1._rk/nm1
    else if (x>1.0) then  !-- Lentz's algorithm
       b=x+n
       c=1.0_rk/FPmin
       d=1.0_rk/b
       h=d
       do i=1, maxIterations
          a=-i*(nm1+i)
          b=b+2.0_rk
          d=1._rk/(a*d+b)
          c=b+a/c
          del=c*d
          h=h*del
          if (abs(del-1.0_rk)<epsilon) then
             expInt=h*exp(-x)
             return
          end if
       end do
       stop 'continued fraction failed in expInt'
    else !-- Evaluate series
       if (nm1/=0) then
          expInt=1./nm1
       else
          expInt=-log(x)-Euler
       end if
       fact=1.0_rk
       do i=1, maxIterations
          fact=-fact*x/i
          if (i/=nm1) then
             del=-fact/(i-nm1)
          else
             psi=-Euler     !-- compute Psi(n)
             do ii=1, nm1
                psi=psi+1.0_rk/ii
             end do
             del=fact*(-log(x)+psi)
          end if
          expInt=expInt+del
          if (abs(del)<abs(expInt)*epsilon) return
       end do
       stop 'series failed in expInt'
    end if
    return
  end function expInt


  !-------------------------------------------------------------------
  !                       FUNCTION int2char
  !-------------------------------------------------------------------
  ! Takes an integer as input and returns the corresponding string.
  ! The 'length' of number must be less or equal than 20.
  !-------------------------------------------------------------------
  function int2char (number)

    integer, intent(in) :: number
    integer, parameter :: length = 20
    character(len=length) :: int2char

    integer :: x, i, j

    integer, dimension(length) :: y !-- array containing each digit in number

    i=length
    y=0
    x = number
    
    do while (x>0)
       y(i) = mod(x,10)
       x=(x-y(i))/10
       i=i-1
       if (i==0) stop 'should not be there!'
    end do

    i=1
    do while (y(i)==0)
       i=i+1
    end do

    int2char = ''
    do j=length, i, -1
       int2char = char(48+y(j))//int2char
    end do

    return

  end function int2char


  !-------------------------------------------------------------------
  !                  subroutine SimpsonIntegration
  !-------------------------------------------------------------------
  ! Computes the integral of a function 'fi' using the Simpson's rule.
  !-------------------------------------------------------------------
  subroutine SimpsonIntegration (n,h,fi,s)
    integer, intent(in)  :: n
    real(rk), intent(in) :: h
    real(rk), intent(in), dimension(n) :: fi
    real(rk), intent(out) :: s
    integer :: i
    real(rk) :: s0, s1, s2    
    s=0.0_rk
    s0=0.0_rk
    s1=0.0_rk
    s2=0.0_rk
    do i=2, n-1, 2
      s1=s1+fi(i-1)
      s0=s0+fi(i)
      s2=s2+fi(i+1)
    enddo
    s=h*(s1+4.0_rk*s0+s2)/3.0_rk
    ! if n is even, add the last slice separately
    if (mod(n,2).eq.0) s=s+h*(5.0_rk*fi(n)+8.0_rk*fi(n-1)-fi(n-2))/12.0_rk
    return
  end subroutine SimpsonIntegration




!-----------------------------------------------------------------------------
!-- interpolation routines (from SLATEC) used in Dave's code
!-----------------------------------------------------------------------------

      subroutine qspl1d2 (n, x, f, w, ij, y, tab)
      real y
      integer ij, n
!     .. array arguments ..
      real f(*), tab(3), w(*), x(*)
!     .. local scalars ..
      real a, b, c, flk
      integer i, k1, mflag, mi
!     .. external subroutines ..
!      external qinterv
!
!     locate y in the x table.
!
      call qinterv(x,n-1,y,i,mflag)
      mi = (i-1)*ij + 1
      k1 = mi + ij
      flk = x(i+1) - x(i)
!
!     calculate f(y).
!
      a = (w(mi)*(x(i+1)-y)**3+w(k1)*(y-x(i))**3)/(6.0*flk)
      b = (f(k1)/flk-w(k1)*flk/6.0)*(y-x(i))
      c = (f(mi)/flk-flk*w(mi)/6.0)*(x(i+1)-y)
      tab(1) = a + b + c
!
!     calculate the first derivative at y.
!
      a = (w(k1)*(y-x(i))**2-w(mi)*(x(i+1)-y)**2)/(2.0*flk)
      b = (f(k1)-f(mi))/flk
      c = flk*(w(mi)-w(k1))/6.0
      tab(2) = a + b + c
!
!     calculate the second derivative at y.
!
      tab(3) = (w(mi)*(x(i+1)-y)+w(k1)*(y-x(i)))/flk
      return
      end subroutine qspl1d2


      subroutine qinterv (xt, lxt, x, ileft, mflag)
!     .. scalar arguments ..
      real x
      integer ileft, lxt, mflag
!     .. array arguments ..
      real xt(*)
!     .. local scalars ..
      integer ihi, ilo, istep, middle
!     .. save statement ..
      save ilo
!     .. data statements ..
      data ilo /1/
      ihi = ilo + 1
      if (ihi .ge. lxt) then
        if (x .ge. xt(lxt)) go to 220
        if (lxt .le. 1) go to 200
        ilo = lxt - 1
        go to 110
      endif
      if (x .ge. xt(ihi)) go to 150
  110 if (x .ge. xt(ilo)) go to 210
!
!     now x .lt. xt(ihi).  find lower bound
!
      istep = 1
  130 ihi = ilo
      ilo = ihi - istep
      if (ilo .le. 1) then
        ilo = 1
        if (x .lt. xt(1)) go to 200
        go to 180
      endif
      if (x .ge. xt(ilo)) go to 180
      istep = istep*2
      go to 130
!
!     now x .ge. xt(ilo).  find upper bound
!
  150 istep = 1
  160 ilo = ihi
      ihi = ilo + istep
      if (ihi .ge. lxt) go to 170
      if (x .lt. xt(ihi)) go to 180
      istep = istep*2
      go to 160
  170 if (x .ge. xt(lxt)) go to 220
      ihi = lxt
!
!     now xt(ilo) .le. x .lt. xt(ihi).  narrow the interval
!
  180 middle = (ilo+ihi)/2
      if (middle .eq. ilo) go to 210
!
!     note.  it is assumed that middle = ilo in case ihi = ilo+1
!
      if (x .lt. xt(middle)) then
        ihi = middle
      else
        ilo = middle
      endif
      go to 180
!
!     set output and return
!
  200 mflag = -1
      ileft = 1
      return
  210 mflag = 0
      ileft = ilo
      return
  220 mflag = 1
      ileft = lxt
      return
      end subroutine qinterv

end module pfnsLosAlamosModel
