module ripl2Masses

  implicit none

  private

  !-------------------- << PUBLIC SUBROUTINES & OBJECTS >> -------------------

  public :: ripl2_readMassesHFB         !-- [subroutine]
  public :: ripl2_readMassesFRDM        !-- [subroutine]

  public :: ripl2_masses                !-- [(Z,A) array]

  !---------------------------------------------------------------------------

  !-- set up accuracy of computation here
  integer, parameter :: rk = selected_real_kind(15,307)

  !-- PATH to data directory
  character(len=*), parameter :: dataDir = "data/"

  !-- masses obtained with HFB calculations (Goriely, 2002)
  character(len=*), parameter :: masses_hfb = trim(dataDir)//'mass-hfb02.dat'

  !-- masses obtained with FRDM calculations (Moller, 1995)
  character(len=*), parameter :: masses_frdm = trim(dataDir)//'mass-frdm95.dat'

  !---------------------------------------------------------------------------
  !-- new type describing nuclear ground-state properties based on HFB calc.
  !---------------------------------------------------------------------------  
  type gshfb
     character(len=2) :: symbol  !-- element 2 letters denomination
     integer          :: flag    !-- flag to specify the nature of the 
                                 !-- given experimental mass
     real(rk) :: Mexp  !-- experimental or recommended atomic mass excess (MeV)
     real(rk) :: Mth   !-- calculatted HFB-2 atomic mass excess (MeV)
     real(rk) :: beta2 !-- quadrupole deformation
     real(rk) :: beta4 !-- hexadecapole deformation
     real(rk) :: rhon  !-- amplitude neutron density distribution
     real(rk) :: rn    !-- radius of neutron density distribution
     real(rk) :: an    !-- diffuseness neutron density distribution
     real(rk) :: rhop  !-- amplitude proton density distribution
     real(rk) :: rp    !-- radius of proton density distribution
     real(rk) :: ap    !-- diffuseness proton density distribution
  end type gshfb


  !---------------------------------------------------------------------------
  !-- new type describing the ground-state properties based on the FRDM model
  !---------------------------------------------------------------------------
  type gsfrdm
     character(len=2) :: El !-- element symbol
     integer :: flag !-- flag corresponding to 
                     !-- 0 if no experimental data available
                     !-- 1 for a mass excess recommended by Audi&Wapstra (1995)
                     !-- 2 for a measured mass from Audi&Wapstra (1995)
     real(rk) :: Mexp !-- exp. or recommended atomic mass excess in MeV of 
                      !-- Audi&Wapstra (1995)
     real(rk) :: Mth  !-- calculated FRDM atomic mass excess in MeV 
     real(rk) :: Emic !-- calculated FRDM microscopic energy in MeV 
     real(rk) :: beta2 !-- calc. quadrupole deformation of the nuclear gs
     real(rk) :: beta3 !-- calc. octupole deformation of the nuclear gs
     real(rk) :: beta4 !-- calc. hexadecapole deformation of the nuclear gs
     real(rk) :: beta6 !-- calc. hexacontatetrapole deformation of the nuclear gs
  end type gsfrdm

  integer, parameter :: Zmax = 150 !-- max. value for the charge number 
  integer, parameter :: Amax = 500 !-- max. value for the mass number

  integer :: eof = 0 !-- test end of file
  logical :: there     !-- boolean variable to check if input file is present

  !-- Array of type(gshfb) that is used to retrieve all information from the file
  type(gshfb), dimension(0:Zmax,0:Amax), target :: groundStatesHFB

  !-- Array of type(gsfrdm) that is used to retrieve all data from file
  type(gsfrdm), dimension(0:Zmax,0:Amax), target :: groundStatesFRDM

  !-- array of excess mass values for the whole (A,Z) range
  !-- NOTE that this array starts at (0,0) and not (1,1)!
  real(rk), dimension(0:Zmax,0:Amax) :: ripl2_masses = 0.



!-------
contains
!-------

  !-------------------------------------------------------------------
  ! Read masses from HFB calculations by Goriely (2002) and store them 
  ! in the general 'ripl2_masses' array.
  !-------------------------------------------------------------------
  subroutine ripl2_readMassesHFB
    integer :: i,j !-- dummy loop indices
    !-- read the ground-states properties from the RIPL2 data file
    call readGroundStatePropertiesHFB
    do i=0, Zmax
       do j=0, Amax
          ripl2_masses(i,j) = groundStatesHFB(i,j)%Mth !-- theoretical mass
       end do
    end do
  end subroutine ripl2_readMassesHFB


  !-------------------------------------------------------------------
  ! Read masses from FRDM (Finite Range Droplet Model) calculations 
  ! by Moller & Nix.
  !-------------------------------------------------------------------
  subroutine ripl2_readMassesFRDM
    integer :: i,j !-- dummy loop indices
    !-- read the ground-states properties from the RIPL2 data file
    call readGroundStatePropertiesFRDM
    do i=0, Zmax
       do j=0, Amax
          ripl2_masses(i,j) = groundStatesFRDM(i,j)%Mexp !-- exp. mass
       end do
    end do
  end subroutine ripl2_readMassesFRDM



  !--------------------------------------------------------------------
  !                SUBROUTINE readGroundStatePropertiesHFB
  !--------------------------------------------------------------------
  ! Read a compilation of the ground-state properties of nuclei 
  ! that have been obtained with Hartree-Fock-Bogolyubov calculations
  ! by S.Goriely et al. [RIPL2]
  !
  !
  ! Each record of the file contains :
  !
  ! Z    : charge number
  ! A    : mass number
  ! El   : element symbol
  ! fl   : flag corresponding to 0 if no experimental data available
  !                              1 for a mass excess recommended by 
  !                                Audi&Wapstra (1995)
  !                              2 for a measured mass from 
  !                                Audi&Wapstra (1995)
  ! Mexp : experimental or recommended atomic mass excess in MeV of 
  !        Audi&Wapstra (1995)
  ! Mth  : calculated HFB-2 atomic mass excess in MeV 
  ! beta2: calculated quadrupole deformation of the nuclear ground-state
  ! beta4: calculated hexadecapole deformation of the nuclear ground-state
  ! rhon : calculated amplitude of the neutron density distribution in fm^-3 
  ! rn   : calculated radius of the neutron density distribution in fm 
  ! an   : calculated diffuseness of the neutron density distribution in fm 
  ! rhop : calculated amplitude of the proton density distribution in fm^-3 
  ! rp   : calculated radius of the proton density distribution in fm 
  ! ap   : calculated diffuseness of the proton density distribution in fm 
  !
  ! The corresponding FORTRAN format is (2i4,1x,a2,1x,i1,2f10.3,2f8.3,6f9.4)
  !
  !--------------------------------------------------------------------
  subroutine readGroundStatePropertiesHFB

    integer :: i !-- dummy loop index

    integer          :: Z,A !-- Z,A of the element g.s. properties read in
    character(len=2) :: El
    integer          :: fl
    real(rk)         :: Mexp,Mth,beta2,beta4,rhon,rn,an,rhop,rp,ap

    !-- initialize the array of ground states
    groundStatesHFB(:,:) = gshfb('**',-1,0,0,0,0,0,0,0,0,0,0)

    !*************************************************************************
    open (unit=10, file=masses_hfb, action='read', status='old')
    !-- skip four comment (#) lines
    do i=1, 4
       read(10,*)
    end do
    !-- read entries until the end of the file
    do while (eof==0)
       read(10,100,iostat=eof) Z, A, El, fl, Mexp, Mth, beta2, beta4, &
            rhon, rn, an, rhop, rp, ap
       groundStatesHFB(Z,A) = gshfb(El,fl,Mexp,Mth,beta2,beta4,rhon,&
            rn,an,rhop,rp,ap)
    end do
    close (10)
100 format(2i4,1x,a2,1x,i1,2f10.3,2f8.3,6f9.4)
    !*************************************************************************

  end subroutine readGroundStatePropertiesHFB
  

  !---------------------------------------------------------------------------
  !                 SUBROUTINE readGroundStatePropertiesFRDM
  !---------------------------------------------------------------------------
  !                          mass-frdm95.dat
  !
  !  Compilations of ground state properties based on the FRDM model
  !	     (provided by S. Goriely on October 5, 2000)
  !  ***************************************************************
  !
  !	     FRDM PREDICTION OF GROUND-STATES PROPERTIES
  !
  !			 P. Moller, J. R. Nix
  !		 Los Alamos National Laboratory (USA)
  !		     W. D. Myers, W. J. Swiatecki
  !		  Lawrence Berkeley Laboratory (USA)
  !
  !
  ! Content
  ! -------
  ! Predictions of the ground state properties obtained within the Finite
  ! Range Droplet Model (FRDM) [1]. The atomic mass excesses and nuclear
  ! ground-state deformations are tabulated for 8979 nuclei ranging from
  ! O-16 to A=339. Calculations are based on the finite-range droplet
  ! macroscopic model and the folded-Yukawa single-particle microscopic
  ! model.  Relative to the 1981 FRDM mass table [2] the current results
  ! are obtained with an improved macroscopic model, an improved pairing
  ! model with a new form for the effective-interaction pairing gap, and
  ! minimization of the ground-state energy with respect to additional
  ! shape degrees of freedom. The values of only nine constants are
  ! determined directly from a least-squares adjustment to the
  ! ground-state masses of 1654 nuclei ranging from O-16 to [Z=106, A=263]
  ! and to 28 fission-barrier heights. The error of the mass model is
  ! 0.669 MeV for the entire region of nuclei considered, but is only
  ! 0.448 MeV for the region N greater than or equal to 65. The table
  ! includes the experimental masses when available, the calculated
  ! masses, the microscopic correction and deformation parameters.  The
  ! microscopic correction Emic corresponds to the difference between the
  ! total binding energy and the spherical macroscopic (droplet) energy.
  !
  !
  ! Format
  ! ------
  ! Each record of the file contains:
  !
  !   Z    : charge number
  !   A    : mass number
  !   El   : element symbol
  !   fl   : flag corresponding to 0 if no experimental data available
  !                                1 for a mass excess recommended by 
  !                                  Audi&Wapstra (1995)
  !                                2 for a measured mass from 
  !                                  Audi&Wapstra (1995)
  !   Mexp : experimental or recommended atomic mass excess in MeV of 
  !          Audi&Wapstra (1995)
  !   Mth  : calculated FRDM atomic mass excess in MeV 
  !   Emic : calculated FRDM microscopic energy in MeV 
  !   beta2: calculated quadrupole deformation of the nuclear ground-state
  !   beta3: calculated octupole deformation of the nuclear ground-state
  !   beta4: calculated hexadecapole deformation of the nuclear ground-state
  !   beta6: calculated hexacontatetrapole deformation of the nuclear 
  !          ground-state
  !
  ! The corresponding FORTRAN format is (2i4,1x,a2,1x,i1,3f10.3,4f8.3)
  !
  !
  ! References
  ! ----------
  ! [1] P. Moller, J. R. Nix, W. D. Myers, W. J. Swiatecki (1995) At. Data 
  !     and Nucl.Data Tables 59, 185.
  ! [2] P. Moller, J. R. Nix (1981) At. Data and Nucl. Data Tables 26, 165.
  ! [3] G. Audi and A. H. Wapstra (1995) Nucl. Phys. A595, 409.
  !
  !---------------------------------------------------------------------------
  subroutine readGroundStatePropertiesFRDM

    integer :: i !-- dummy loop index

    integer          :: Z,A !-- Z,A of the element g.s. properties read in
    character(len=2) :: El
    integer          :: fl
    real(rk)         :: Mexp,Mth,Emic,beta2,beta3,beta4,beta6

    !-- initialize the array of ground states
    groundStatesFRDM(:,:) = gsfrdm('**',-1,0,0,0,0,0,0,0)

    !*************************************************************************
    open (unit=10, file=masses_frdm, action='read', status='old')
    !-- skip four comment (#) lines
    do i=1, 4
       read(10,*)
    end do
    !-- read entries until the end of the file
    eof=0
    do while (eof==0)
       read(10,100,iostat=eof) &
            Z, A, El, fl, Mexp, Mth, Emic, beta2, beta3, beta4, beta6
       groundStatesFRDM(Z,A) = &
            gsfrdm(El,fl,Mexp,Mth,Emic,beta2,beta3,beta4,beta6)
    end do
    close (10)
100 format(2i4,1x,a2,1x,i1,3f10.3,4f8.3)
    !*************************************************************************


  end subroutine readGroundStatePropertiesFRDM



end module ripl2Masses
